/*
   CIPE - encrypted IP over UDP tunneling

   ciped.c - the user mode driver

   Copyright 1996 Olaf Titz <olaf@bigred.inka.de>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version
   2 of the License, or (at your option) any later version.
*/
/* $Id: ciped.c,v 1.16 1998/04/07 20:08:51 olaf Exp $ */

#define CTLDIR "/etc/cipe"

// #include <linux/in.h>       /* from 5.3e */
#include <errno.h>          /* from 5.6  */
#include <fcntl.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/signal.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/wait.h>
#include "ciped.h"
#include "ioctl.h"

static const char daemon_version[]=VERSION;

void usage(const char *p)
{
    fprintf(stderr, "usage: %s [-o file ] [options] device=name\n", p);
    exit(1);
}

/* FIXME: places marked *#* have byteorder conversion missing */
/* This is corrected in version 3 protocol */

/*** The global options structure ***/

struct options {
    const char * const oname;
    const char otyp;
    union {
	int ovint;
	char *ovstr;
	struct sockaddr_in *ovsaddr;
    } ov;
} opts[] = {
#define Opt(x,t) { #x, T##t, {0} },
#define Tbool	0
#define Tint	1
#define Tstr	2
#define Taddr	3
#define Tuaddr	4
#define Ttaddr	5

#define Odevice		0  /* name of the CIPE device, e.g. cip3b0 */
Opt(device,str)
#define Odebug		1  /* Debug mode - don't detach, verbose messages */
Opt(debug,bool)
#define Oipaddr		2  /* This CIPE device's IP address */
Opt(ipaddr,addr)
#define Optpaddr	3  /* The peer CIPE device's IP address */
Opt(ptpaddr,addr)
#define Omtu		4  /* Device MTU */
Opt(mtu,int)
#define Ometric		5  /* Device metric */
Opt(metric,int)
#define Ome		6  /* My UDP address */
Opt(me,uaddr)
#define Opeer		7  /* The peer's UDP address,
			      i.e. where to send packets */
Opt(peer,uaddr)
#define Okey		8  /* The interface secret key
			      - set via options file only */
Opt(key,str)
#define Onokey		9  /* Don't encrypt */
Opt(nokey,bool)
#define Osocks		10  /* SOCKS5 server (ip:port) */
Opt(socks,taddr)
#define Otokxc		11  /* Timeout for key exchange (secs) */
Opt(tokxc,int)
#define Otokey		12  /* Dynamic key lifetime (secs) */
Opt(tokey,int)
#define Oipup		13  /* Name of the ip-up script */
Opt(ipup,str)
#define Oipdown		14  /* Name of the ip-down script */
Opt(ipdown,str)
#define Oarg		15  /* Extra argument to ip-up script */
Opt(arg,str)
#define Octtl		16  /* TTL value for carrier */
Opt(cttl,int)
#define Omaxerr		17  /* Maximum ECONNREFUSED tolerance */
Opt(maxerr,int)

#undef Opt
{ NULL, 0, {0} }};

#define OI(x)      (opts[O##x].ov.ovint)
#define OS(x)      (opts[O##x].ov.ovstr)
#define OA(x)      (opts[O##x].ov.ovsaddr)
#define OAaddr(x)  (OA(x)->sin_addr)
#define OAport(x)  (OA(x)->sin_port)

/* Statistics */
struct kstat {
    int rreq, req, ind, indb, ack, ackb, bogus;
} ks = { 0, 0, 0, 0, 0, 0, 0 };


/*** Option parser ***/

/* Resolve a host:port address */
int getaddr(char *c, struct sockaddr_in *sa, const char *prot)
{
    struct hostent *h;
    struct servent *s;
    char *q=strchr(c, ':');

    sa->sin_family=AF_INET;
    sa->sin_port=0;
    sa->sin_addr.s_addr=INADDR_ANY;
    if (q) {
	if (!prot) {
	    fprintf(stderr, "raw IP address has no port number\n");
	    return -1;
	}
	*q++='\0';
	if (!(sa->sin_port=htons(atoi(q)))) {
	    if ((s=getservbyname(q, prot)))
		sa->sin_port=s->s_port;
	    else {
		fprintf(stderr, "port '%s' invalid/not found\n", q);
		return -1;
	    }
	}
    }
    if ((sa->sin_addr.s_addr=inet_addr(c))==-1) {
	if ((h=gethostbyname(c)))
	    memcpy(&sa->sin_addr.s_addr, h->h_addr_list[0],
		   sizeof(sa->sin_addr.s_addr));
	else {
	    fprintf(stderr, "host '%s' invalid/not found\n", c);
	    return -1;
	}
    }
    return 0;
}

/* Set an option */
int setopt(const char *n, char *v)
{
    struct options *o=opts;
    static const char * const protos[]={NULL, "udp", "tcp"};

    while (o->oname) {
	if (!strcmp(o->oname, n)) {
	    if ((!v) && (o->otyp!=Tbool)) {
		fprintf(stderr, "missing value for %s\n", n); return -1;
	    }
	    switch(o->otyp) {
	    case Tbool:
		o->ov.ovint=1; break;
	    case Tint:
		o->ov.ovint=atoi(v); break;
	    case Tstr:
		if (!(o->ov.ovstr=strdup(v))) {
		    fprintf(stderr, "setopt: out of memory\n"); return -1;
		}
		break;
	    case Taddr:
	    case Tuaddr:
	    case Ttaddr:
		if (!(o->ov.ovsaddr=malloc(sizeof(*o->ov.ovsaddr)))) {
		    fprintf(stderr, "setopt: out of memory\n"); return -1;
		}
		if (getaddr(v, o->ov.ovsaddr, protos[o->otyp-Taddr])<0) {
		    return -1;
		}
		break;
	    default:
		fprintf(stderr, "internal error: %d\n", __LINE__); return -1;
	    }
	    return 0;
	}
	++o;
    }
    fprintf(stderr, "unknown option: %s\n", n);
    return -1;
}

/* Set options from command line */
void setopt_cmdline(int argc, char *argv[])
{
    int i;
    char *q;
    for (i=0; i<argc; ++i) {
	if ((q=strchr(argv[i], '=')))
	    *q++='\0';
	if (!strcmp(argv[i], "key")) {
	    fprintf(stderr, "Refused to set key from command line!\n");
	    continue;
	}
	(void) setopt(argv[i], q);
    }
}

/* Set options from file, report file not found if (v) */
void setopt_file(const char *n, int v)
{
    int d;
    FILE *f;
    char b[128];
    char *p, *q;
    struct stat s;

    if ((d=open(n, O_RDONLY))<0) {
	if (v)
	    perror("setopt_file: open");
	return;
    }
    if (fstat(d, &s)<0) {
	perror("setopt_file: stat");
	return;
    }
    if (!S_ISREG(s.st_mode) || (s.st_mode&(S_IRWXG|S_IRWXO)) ||
	(s.st_uid!=0)) {
	fprintf(stderr, "file %s has incorrect permissions\n", n);
	return;
    }

    if (!(f=fdopen(d, "r"))) {
	if (v)
	    perror("setopt_file: fdopen");
	return;
    }
    while (fgets(b, sizeof(b), f)) {
	p=b;
	if ((q=strchr(p, '\n')))
	    *q='\0';
	while ((*p) && strchr(" \t", *p))
	    ++p;
	if ((!*p) || strchr("#;:!", *p))
	    continue;
	if ((q=strpbrk(p, " \t="))) {
	    while ((*q) && strchr(" \t=", *q))
		*q++='\0';
	    if (!*q)
		q=NULL;
	}
	(void) setopt(p, q);
    }
    fclose(f);
}

/* Print out all options */
void dumpopt(void)
{
    struct options *o;
    for (o=opts; o->oname; ++o) {
	printf("%s ", o->oname);
	switch(o->otyp) {
	case Tbool: printf("%s\n", o->ov.ovint?"yes":"no"); break;
	case Tint:  printf("%d\n", o->ov.ovint); break;
	case Tstr:  printf("%s\n", o->ov.ovstr?o->ov.ovstr:"(none)"); break;
	case Taddr:
	    if (o->ov.ovsaddr)
		printf("%s\n", inet_ntoa(o->ov.ovsaddr->sin_addr));
	    else
		printf("(none)\n");
	    break;
	case Tuaddr:
	case Ttaddr:
	    if (o->ov.ovsaddr)
		printf("%s:%d\n", inet_ntoa(o->ov.ovsaddr->sin_addr),
		       ntohs(o->ov.ovsaddr->sin_port));
	    else
		printf("(none)\n");
	    break;
	default: printf("internal error: %d\n", __LINE__); break;
	}
    }
}

/*** Logging ***/

void log(int lev, const char *s)
{
    if (OI(debug))
	fprintf(stderr, s);
    else
	syslog(lev, s);
}

void log2(int lev, const char *f, const char *s)
{
    if (OI(debug))
	fprintf(stderr, f, s);
    else
	syslog(lev, f, s);
}

/* perror replacement */
void logerr(int lev, const char *s)
{
    if (OI(debug))
	perror(s);
    else
	syslog(lev, "%s: %s", s, strerror(errno));
}

/* missing option */
void argmiss(const char *s)
{
    if (OI(debug))
	fprintf(stderr, "missing argument: %s\n", s);
    else
	syslog(LOG_ERR, "missing argument: %s", s);
}

/* print if debugging */
#define dprintf0(f)    if(OI(debug)){printf(f);}
#define dprintf1(f,a)  if(OI(debug)){printf(f,a);}


/*** Initialize, open and attach the CIPE device. ***/

int opendev(void)
{
    int f, i;
    struct siocsifcippar p;
    struct siocsifcipkey k;
    struct siocsifcipatt a;
    struct ifreq r;

#define amiss(s) { argmiss(s); goto error; }
#define err(s)   { logerr(LOG_ERR, s); goto error; }

    if (!OS(device)) {
	argmiss("device");
	return -1;
    }

    /* Initialize, bind and connect the socket */
    if ((f=socket(AF_INET, SOCK_DGRAM, 0))<0) {
	logerr(LOG_ERR, "opendev: socket");
	return -1;
    }
    i=65536; /* we need much to receive fast */
    if (setsockopt(f, SOL_SOCKET, SO_RCVBUF, &i, sizeof(i)))
	/* not fatal */
	logerr(LOG_NOTICE, "opendev: setsockopt");
    if (!OA(me)) {
	if (!(OA(me)=malloc(sizeof(*OA(me)))))
	    err("opendev: malloc");
	OA(me)->sin_family=AF_INET;
	OAaddr(me).s_addr=INADDR_ANY;
	OAport(me)=0;
    }
    if (bind(f, (struct sockaddr *)OA(me), sizeof(*OA(me)))<0)
	err("opendev: bind");
    if (OA(peer)) {
	if (connect(f, (struct sockaddr *)OA(peer), sizeof(*OA(peer)))<0)
	    err("opendev: bind");
    } else
	amiss("peer");
    i=sizeof(*OA(me));
    if (getsockname(f, (struct sockaddr *)OA(me), &i)<0)
	/* not fatal */
	logerr(LOG_NOTICE, "opendev: getsockname");

    /* Set the CIPE operation parameters */
    memset(&p, 0, sizeof(p));
    if (OA(socks)) {
	p.sockshost=OAaddr(socks).s_addr;
	p.socksport=OAport(socks);
    }
    p.tmo_keyxchg=OI(tokxc);
    p.tmo_keylife=OI(tokey);
    p.mayclear=OI(nokey);
    p.cttl=OI(cttl);
    if (ioctl_setpar(f, OS(device), &p)<0)
	err("opendev: setpar");

    /* Set the key */
    if (OS(key)) {
	unsigned char *c=OS(key);
	unsigned char *x=(unsigned char *)&k.thekey;
	i=0;
	k.which=KEY_STATIC;
	/* primitive hex decoder */
	while(*c) {
	    x[i]=((*c)<<4)&0xF0;
	    *c++='\0'; /* clear after use */
	    if (!*c) break;
	    x[i++]|=(*c)&0x0F;
	    *c++='\0';
	    if (i>=userKeySize) break;
	}
	if (ioctl_setkey(f, OS(device), &k)<0)
	    err("opendev: setkey");
	memset(&k, 0, sizeof(k));
    } else {
	if (!OI(nokey))
	    amiss("key");
    }

    /* Perform attach */
    a.fd=f;
    if (ioctl_attach(f, OS(device), &a)<0)
	err("opendev: attach");

    /* Perform ifconfig */
    strcpy(r.ifr_name, OS(device));
    if (OA(ipaddr)) {
	memcpy(&r.ifr_addr, OA(ipaddr), sizeof(*OA(ipaddr)));
	if (ioctl(f, SIOCSIFADDR, &r)<0)
	    err("opendev: SIOCSIFADDR");
    } else
	amiss("ipaddr");
    if (OA(ptpaddr)) {
	memcpy(&r.ifr_dstaddr, OA(ptpaddr), sizeof(*OA(ptpaddr)));
	if (ioctl(f, SIOCSIFDSTADDR, &r)<0)
	    err("opendev: SIOCSIFDSTADDR");
    }
    if (OI(mtu)) {
	r.ifr_mtu=OI(mtu);
	if (ioctl(f, SIOCSIFMTU, &r)<0)
	    err("opendev: SIOCSIFMTU");
    }
    if (OI(metric)) {
	r.ifr_metric=OI(metric);
	if (ioctl(f, SIOCSIFMETRIC, &r)<0)
	    err("opendev: SIOCSIFMETRIC");
    }
    r.ifr_flags=IFF_UP|IFF_RUNNING|IFF_NOARP|IFF_NOTRAILERS|IFF_POINTOPOINT;
    if (ioctl(f, SIOCSIFFLAGS, &r)<0)
	err("opendev: SIOCSIFFLAGS");

    return f;

 error:
    close(f);
    return -1;

#undef amiss
#undef err
}

/* Close the device. */

void closedev(int f)
{
    struct ifreq r;
    strcpy(r.ifr_name, OS(device));
    if (ioctl(f, SIOCGIFFLAGS, &r)<0)
	logerr(LOG_ERR, "closedev: SIOCGIFFLAGS");
    r.ifr_flags&=~(IFF_UP|IFF_RUNNING);
    if (ioctl(f, SIOCSIFFLAGS, &r)<0)
	logerr(LOG_ERR, "closedev: SIOCSIFFLAGS");
    close(f);
}

/*** Open a SOCKS5 connection ***/

int opensocks(void) {
    struct sockaddr_in a;
    int fd=socks5_open(OA(socks));
    if (fd>=0) {
	/* 3rd par of socks5_cmd is in: my address, out: relayer */
	a=*OA(me);
	fd=socks5_cmd(fd, S5_UDP, &a);
	/* Swap peer and socks addresses.
	   Use the SOCKS relayer as "peer" and the real peer as "socks"
	   parameter. See output.c why. */
	*OA(socks)=*OA(peer);
	*OA(peer)=a;
    }
    return fd;
}

/* Check whether the SOCKS5 server has closed the connection */

const char *eofsocks_str[] = {"no error", "exception", "read error", "EOF"};

int eofsocks(int fd) {
    fd_set ifd, efd;
    struct timeval t={0,0};
    FD_ZERO(&ifd);
    FD_ZERO(&efd);
    FD_SET(fd, &ifd);
    FD_SET(fd, &efd);
    if (select(fd+1, &ifd, NULL, &efd, &t)>0) {
	if (FD_ISSET(fd, &efd)) {
	    return 1;
	}
	if (FD_ISSET(fd, &ifd)) {
	    int e;
	    char c;
	    if ((e=read(fd, &c, 1))<=0) {
		return e<0 ? 2 : 3;
	    }
	}
    }
    return 0;    
}

/*** Signal handler ***/

sig_atomic_t gotsig=0;

void sighand(int s)
{
    if (OI(debug)) {
	printf("Signal %d\n", s);
    } else {
	syslog(LOG_NOTICE, "got signal %d", s);
    }
    gotsig=s;
}

void sigignore(int s)
{
    /* Dummy */
}

void setsig(int sig, void (*fun)(int))
{
  struct sigaction sa;
  sa.sa_handler = fun;
  sa.sa_flags = 0;
  sigemptyset(&sa.sa_mask);
  sigaction(sig, &sa, NULL);
}

/*** The key exchange procedure ***/

#ifdef VER_CRC32
typedef unsigned long crc;
#define CRCFORM "%08lX"
#else
typedef unsigned short crc;
#define CRCFORM "%04X"
#endif

#define REM 	256	/* How many keys to remember */

/* Cache last valid key CRCs. */

static int dupcheck(crc c)
{
    static crc ca[REM];
    static int caf=0, cal=0;
    int i;

    for (i=cal; i<caf; ++i)
	if (ca[i%REM]==c)
	    return 1;
    if (caf-cal>=REM)
	if (++cal>REM) {
	    caf-=REM;
	    cal-=REM;
	}
    ca[caf%REM]=c;
    ++caf;
    return 0;
}

int errcnt;

int kxchg(int f, int r)
{
    int x=0, e=0;
    unsigned char kx[KEYXCHGBLKMIN];
    #define kx_typ       (kx[0])
    #define kx_nkind_key (kx+1)
#ifdef VER_CRC32
    #define kx_nkind_crc (*(unsigned long *)(kx+1+userKeySize))
    #define kx_nkack_crc (*(unsigned long *)(kx+1))
#else
    #define kx_nkind_crc (*(unsigned short *)(kx+1+userKeySize))
    #define kx_nkack_crc (*(unsigned short *)(kx+1))
#endif
    static crc scrc=0;
    static unsigned char skey[userKeySize];

    if (recv(f, &kx, sizeof(kx), 0)<=0) {
	if (errno==EAGAIN || errno==EINTR)
	     /* This is normal, would be even after select */
	    return 0;
	logerr(LOG_INFO, "kxchg: recv");
	if (errno==ENXIO)
	    /* Device was closed */
	    return -1;
	/* allow for some amount of ECONNREFUSED etc. */
	return (OI(maxerr)<0 ? 0 : --errcnt);
    }
    errcnt=OI(maxerr);
    switch(kx_typ) {
    case NK_RREQ:
	kx_typ=NK_REQ;
	dprintf0("KX: [NK_RREQ] sending NK_REQ\n");
	x=send(f, &kx, sizeof(kx), 0); e=errno;
	++ks.rreq;
	break;
    case NK_REQ:
	kx_typ=NK_IND;
	if (read(r, &skey, userKeySize)!=userKeySize) {
	    logerr(LOG_ERR, "kxchg: read(r)");
	    return -1;
	}
	memcpy(kx_nkind_key, skey, userKeySize);
#ifdef VER_CRC32
	kx_nkind_crc=(scrc=htonl(crc32(skey, userKeySize)));
#else
	kx_nkind_crc=(scrc=block_crc(skey, userKeySize)); /*#*/
#endif
	dprintf1("KX: [NK_REQ] sending NK_IND " CRCFORM "\n", scrc);
	x=send(f, &kx, sizeof(kx), 0); e=errno;
	{
	    struct siocsifcipkey sk;
	    sk.which=KEY_INVAL+KEY_SEND;
	    if (ioctl_setkey(f, OS(device), &sk)<0)
		logerr(LOG_ERR, "setkey");
	    memset(&sk, 0, sizeof(sk));
	}
	++ks.req;
	break;
    case NK_IND:
	if (
#ifdef VER_CRC32
	    crc32(kx_nkind_key, userKeySize)==ntohl(kx_nkind_crc)
#else
	    block_crc(kx_nkind_key, userKeySize)==(kx_nkind_crc) /*#*/
#endif
	    ) {
	    struct siocsifcipkey sk;
	    if (dupcheck(kx_nkind_crc)) {
		dprintf0("KX: [NK_IND] duplicate\n");
		++ks.indb;
	    } else {
		sk.which=KEY_RECV;
		memcpy(&sk.thekey, kx_nkind_key, userKeySize);
		if (ioctl_setkey(f, OS(device), &sk)<0)
		    perror("setkey");
		else {
		    kx_typ=NK_ACK;
		    kx_nkack_crc=kx_nkind_crc; /* XX */
		    dprintf1("KX: [NK_IND] sending NK_ACK " CRCFORM "\n",
			     (kx_nkack_crc)); /*#*/
		    x=send(f, &kx, sizeof(kx), 0); e=errno;
		}
		++ks.ind;
	    }
	    memset(&sk, 0, sizeof(sk));
	} else {
	    dprintf0("KX: [NK_IND] invalid\n");
	    ++ks.indb;
	}
	break;
    case NK_ACK:
	if (scrc==(kx_nkack_crc)) { /*#*/
	    struct siocsifcipkey sk;
	    sk.which=KEY_SEND;
	    memcpy(&sk.thekey, &skey, userKeySize);
	    dprintf1("KX: [NK_ACK] got " CRCFORM "\n", scrc);
	    if (ioctl_setkey(f, OS(device), &sk)<0)
		perror("setkey");
	    ++ks.ack;
	    memset(&sk, 0, sizeof(sk));
	    memset(&skey, 0, sizeof(skey));
	} else {
	    dprintf1("KX: [NK_ACK] unknown " CRCFORM "\n", scrc);
	    ++ks.ackb;
	}
	break;
    default:
	dprintf1("KX: invalid type %02X\n", kx_typ);
	++ks.bogus;
    }
    memset(kx, 0, sizeof(kx));
    if (x<0) {
	errno=e;
	logerr(LOG_WARNING, "kxchg: send");
    }
    return x;
}

/*** Call the ip-up, ip-down scripts ***/

void pspawn(char *s)
{
    int p, c;
    p=getpid();
    if ((c=fork())<0) {
	logerr(LOG_ERR, "pspawn: fork");
	return;
    }
    if (c) {
	(void) wait(NULL); /* exactly one child is active */
    } else {
	/* call <interface> <my-addr> <daemon-pid> <local> <remote> <arg> */
	char bufa[32], bufb[32], bufc[32], bufd[32];
	char *na[]={s, OS(device), bufa, bufb, bufc, bufd, OS(arg), NULL};
	sprintf(bufa, "%s:%d", inet_ntoa(OAaddr(me)), ntohs(OAport(me)));
	sprintf(bufb, "%d", p);
	sprintf(bufc, "%s", inet_ntoa(OAaddr(ipaddr)));
	sprintf(bufd, "%s", inet_ntoa(OAaddr(ptpaddr)));
	close(0); close(1); close(2);
	(void) open("/dev/null", O_RDWR);
	(void) dup(0); (void) dup(0);
	execv(s, na);
	logerr(LOG_WARNING, "pspawn: exec");
	exit(1);
    }
}

/*** Main program ***/

int main(int argc, char *argv[])
{
    int i, ur, fd, sd=-1;
    char *pn=strrchr(argv[0], '/');
    if (pn)
	++pn;
    else
	pn=argv[0];

    if (getuid() || geteuid()) {
	fprintf(stderr, "This program must be run by root.\n");
	exit(1);
	/* DO NOT remove this check */
    }
    /* Snarf options from: standard options file, user-supplied file and
       finally, command line */
    setopt_file(CTLDIR "/options", 0);
    if ((argc>1) && (!strcmp(argv[1], "-o"))) {
	if (argc>2) {
	    setopt_file(argv[2], 1);
	    argc-=2; argv+=2;
	} else {
	    usage(pn);
	}
    }
    setopt_cmdline(--argc, ++argv);

    for (i=getdtablesize()-1; i>2; --i)
	(void) close(i);

    if (!OI(debug)) {
	/* daemonize */
	if ((i=fork())<0) {
	    perror("fork"); exit(1);
	}
	if (i)
	    exit(0);
	openlog(pn, LOG_PID, LOG_DAEMON);
	/* don't let the syslog socket get in the way of the standard fds */
	close(0); close(1); close(2);
	setsid();
	syslog(LOG_INFO, "%s: CIPE daemon vers %s (c) Olaf Titz 1996-1998",
	       OS(device), daemon_version);
    } else {
	printf("%s: CIPE daemon vers %s (c) Olaf Titz 1996-1998\n",
	       OS(device), daemon_version);
	dumpopt();
    }

    if ((ur=open("/dev/urandom", O_RDONLY))<0) {
	logerr(LOG_ERR, "open(/dev/urandom)"); exit(1);
    }
    if (OA(socks)) {
	struct itimerval t;
	if ((sd=opensocks())<0)
	    exit(1);
	/* Since the key exchange socket can't do select(), we make
	   the recv() return EINTR every now and then to get a chance
	   to check the socks5 control socket. */
	setsig(SIGALRM, sigignore);
	t.it_interval.tv_sec = 30;
	t.it_interval.tv_usec = 0;
	t.it_value.tv_sec = 60;
	t.it_value.tv_usec = 0;
	setitimer(ITIMER_REAL, &t, 0);
    }
    if ((fd=opendev())<0)
	exit(1);
    pspawn(OS(ipup) ? OS(ipup) : CTLDIR "/ip-up");
    setsig(SIGHUP, sighand);
    setsig(SIGINT, sighand);
    setsig(SIGTERM, sighand);
    if (!OI(maxerr))
	OI(maxerr)=8; /* compatibility */
    errcnt=OI(maxerr);

    /* Main loop */
    while ((!gotsig) && (kxchg(fd, ur)>=0)) {
	if ((sd>=0) && (i=eofsocks(sd))) {
	    log2(LOG_INFO, "%s on socks connection", eofsocks_str[i]);
	    break;
	}
    }

    closedev(fd);
    if (OI(debug))
	printf("KX stats: rreq=%d, req=%d, ind=%d, indb=%d, ack=%d,"
	       " ackb=%d, unknown=%d\n", ks.rreq, ks.req, ks.ind, ks.indb,
	       ks.ack, ks.ackb, ks.bogus);
    else
	syslog(LOG_INFO, "KX stats: rreq=%d, req=%d, ind=%d, indb=%d, ack=%d,"
	       " ackb=%d, unknown=%d\n", ks.rreq, ks.req, ks.ind, ks.indb,
	       ks.ack, ks.ackb, ks.bogus);
    pspawn(OS(ipdown) ? OS(ipdown) : CTLDIR "/ip-down");
    if (!OI(debug))
	syslog(LOG_INFO, "%s: daemon exiting", OS(device));
    return 0;
}
