
/*
   CIPE - encrypted IP over UDP tunneling

   lite.c - the compression part of the output routine

   Copyright 1997 Allan Latham <alatham@flexsys-group.com>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version
   2 of the License, or (at your option) any later version.
*/

#include <net/ip.h>
#include <net/tcp.h>
#include <linux/if_arp.h>

#include "cipe.h"
#include "liteprototypes.h"

#define EMPTY      0
#define IKNOW      1
#define UKNOW      2
#define IKNOWUKNOW 3
#define NOTFOUND   127
#define OKFLAG     128


/*-----------------------------------------------------------------*/
int lite_get_next_free(struct pairs *pairs, __u32 saddr, __u32 daddr)
{
 int i;

 for (i=0; i < LITE_PAIRS_MAX; i++)
 {
  if ((pairs->saddr[i] == 0) || (pairs->daddr[i] == 0) 
                             || (pairs->ok[i] == EMPTY))
  {
   pairs->saddr[i] = saddr;
   pairs->daddr[i] = daddr;
   pairs->ok[i] = IKNOW;
   return(i);
  }
 }
 return(NOTFOUND);
}

/*-----------------------------------------------------------------*/
int lite_get_pair_from_addresses(struct pairs *pairs, __u32 saddr,
                                                      __u32 daddr, int sr)
{
 __u32 da, sa;
 int i;

 if (sr == LITE_SEND)
 { 
  da = daddr;
  sa = saddr;
 }
 else
 { 
  sa = daddr;
  da = saddr;
 }

 for (i=0; i < LITE_PAIRS_MAX; i++)
 {
  if ((pairs->saddr[i] == sa) && (pairs->daddr[i] == da) 
                              && (pairs->ok[i] != EMPTY))
  {
   return(i);
  }
 }
 return(NOTFOUND);
}

/*-----------------------------------------------------------------*/
void liten(struct cipe *cipe, struct sk_buff *skb)
{
 struct iphdr  oph;      /* Copy of the old IP header */
 struct tcphdr otcph;    /* Copy of the old TCP header (if any) */
 struct ciphdr *ciph;    /* The old IP header (compresed) */
 struct ctcphdr *ctcph;  /* The old TCP header (compresed) */

 unsigned char ttl, pair, version;

 memcpy(&oph, skb->ip_hdr, sizeof(struct iphdr));

 if (oph.ihl != 5)
 {
  cipe->tx_lite_other++;
  return;
 }

 if ((oph.frag_off & htons(IP_MF|IP_OFFSET)) != 0)
 {
  cipe->tx_lite_other++;
  return;
 }

 if ((oph.protocol != IPPROTO_TCP)
  && (oph.protocol != IPPROTO_UDP)
  && (oph.protocol != IPPROTO_ICMP))
 {
  cipe->tx_lite_other++;
  return;
 }

 skb_pull(skb, sizeof(struct iphdr));

 version = LITE_NORMAL;

 if (oph.protocol == IPPROTO_TCP)
 {
  memcpy(&otcph, skb->data, sizeof(struct tcphdr));
  if ((otcph.urg_ptr == 0)
   && (ntohs(otcph.window) < 32768)
   && (otcph.doff == 5)
   && (otcph.ack  == 1)
   && (otcph.res1 == 0)
   && (otcph.res2 == 0)
   && (otcph.urg  == 0)
   && (otcph.rst  == 0)
   && (otcph.syn  == 0)
   && (otcph.fin  == 0))
  {
   version = LITE_COMP;
   ctcph = (struct ctcphdr*)skb_pull(skb, sizeof(struct tcphdr) - CTCPHDRLEN);
   ctcph->source  = otcph.source;
   ctcph->dest    = otcph.dest;
   ctcph->seq     = otcph.seq;
   ctcph->ack_seq = otcph.ack_seq;
   ctcph->check   = otcph.check;
   ctcph->winpsh  = htons((otcph.psh * 32768) + ntohs(otcph.window));
  }
 }

 if (oph.ttl < 16)
 {
  ttl = oph.ttl;
 }
 else
 {
  ttl = 15;
 }

 pair = lite_get_pair_from_addresses(&cipe->pairs, oph.saddr,
                                                 oph.daddr, LITE_SEND);
 if (pair == NOTFOUND)
 {
  pair = lite_get_next_free(&cipe->pairs, oph.saddr, oph.daddr);
 }
 else
 {
  if (cipe->pairs.ok[pair] == IKNOWUKNOW)
  {
   pair += OKFLAG;
  }
 }
 if (pair < OKFLAG)  
 {
  if (pair != NOTFOUND)
  {
   if (cipe->pairs.ok[pair] == UKNOW)
   {
    pair += OKFLAG;
   }
  }
  ciph = (struct ciphdr*)skb_push(skb, CIPHDRLEN);
  memcpy(&ciph->saddr0,&oph.saddr,8);
 }
 else
 {
  ciph = (struct ciphdr*)skb_push(skb, PIPHDRLEN);
  version += LITE_SHORT;
 }
 if (oph.protocol == IPPROTO_TCP)
 {
  ciph->version = version + LITE_TCP;
 }
 else
 {
  if (oph.protocol == IPPROTO_UDP)
  {
   ciph->version = version + LITE_UDP;
  }
  else
  {
   ciph->version = version + LITE_ICMP;
  }
 }

 switch (ciph->version)
 {
  case LITE_ICMP:
    cipe->tx_lite_icmp++;
    break;
  case LITE_ICMP_SHORT:
    cipe->tx_lite_icmp_short++;
    break;
  case LITE_UDP:
    cipe->tx_lite_udp++;
    break;
  case LITE_UDP_SHORT:
    cipe->tx_lite_udp_short++;
    break;
  case LITE_TCP:
    cipe->tx_lite_tcp++;
    break;
  case LITE_TCP_SHORT:
    cipe->tx_lite_tcp_short++;
    break;
  case LITE_TCP_COMP:
    cipe->tx_lite_tcp_comp++;
    break;
  case LITE_TCP_SHORT_COMP:
    cipe->tx_lite_tcp_short_comp++;
    break;
  default:
    cipe->tx_lite_other++;
 }

 ciph->pair = pair;
 ciph->ihl = ttl;
 return;
}

/*-----------------------------------------------------------------*/
void litex(struct cipe *cipe, struct sk_buff *skb, int *length, int *offset)
{
 struct ciphdr ciph;
 struct ctcphdr ctcph;
 struct iphdr *iph;
 struct tcphdr *tcph;
 unsigned char ttl, proto, pair, len;

 memcpy(&ciph, skb->data, CIPHDRLEN);

 switch (ciph.version)
 {
  case LITE_ICMP:
    cipe->rx_lite_icmp++;
    break;
  case LITE_ICMP_SHORT:
    cipe->rx_lite_icmp_short++;
    break;
  case LITE_UDP:
    cipe->rx_lite_udp++;
    break;
  case LITE_UDP_SHORT:
    cipe->rx_lite_udp_short++;
    break;
  case LITE_TCP:
    cipe->rx_lite_tcp++;
    break;
  case LITE_TCP_SHORT:
    cipe->rx_lite_tcp_short++;
    break;
  case LITE_TCP_COMP:
    cipe->rx_lite_tcp_comp++;
    break;
  case LITE_TCP_SHORT_COMP:
    cipe->rx_lite_tcp_short_comp++;
    break;
  default:
    cipe->rx_lite_other++;
 }

 if ((ciph.version == LITE_TCP)
  || (ciph.version == LITE_TCP_SHORT)
  || (ciph.version == LITE_TCP_COMP)
  || (ciph.version == LITE_TCP_SHORT_COMP))
 {
  proto = IPPROTO_TCP;
 }
 else
 {
  if ((ciph.version == LITE_UDP) || (ciph.version == LITE_UDP_SHORT))
  {
   proto = IPPROTO_UDP;
  }
  else
  {
   if ((ciph.version == LITE_ICMP) || (ciph.version == LITE_ICMP_SHORT))
   {
    proto = IPPROTO_ICMP;
   }
   else
   {
    return; 
   }
  }
 }

 ttl = ciph.ihl;
 
 len = 0;

 if ((ciph.version == LITE_TCP) 
  || (ciph.version == LITE_TCP_COMP) 
  || (ciph.version == LITE_UDP) 
  || (ciph.version == LITE_ICMP))
 {
  skb_pull(skb, CIPHDRLEN);
  len -= CIPHDRLEN;
 }
 else
 {
  skb_pull(skb, PIPHDRLEN);
  len -= PIPHDRLEN;
 }

 if ((ciph.version == LITE_TCP_SHORT_COMP)
  || (ciph.version == LITE_TCP_COMP))
 {
  memcpy(&ctcph, skb->data, CTCPHDRLEN);
  skb_pull(skb, CTCPHDRLEN);
  len -= CTCPHDRLEN;
  tcph = (struct tcphdr*)skb_push(skb, sizeof(struct tcphdr));
  len += sizeof(struct tcphdr);
  tcph->source  = ctcph.source;
  tcph->dest    = ctcph.dest;
  tcph->seq     = ctcph.seq;
  tcph->ack_seq = ctcph.ack_seq;
  if (ntohs(ctcph.winpsh) > 32767)
  {
   tcph->psh     = 1;
   tcph->window  = htons(ntohs(ctcph.winpsh) - 32768);
  }
  else
  {
   tcph->psh     = 0;
   tcph->window  = ctcph.winpsh;
  }
  tcph->doff    = 5;
  tcph->ack     = 1;
  tcph->urg_ptr = 0;
  tcph->res1    = 0;
  tcph->res2    = 0;
  tcph->urg     = 0;
  tcph->rst     = 0;
  tcph->syn     = 0;
  tcph->fin     = 0;
  tcph->check   = ctcph.check;
/*
  printk(KERN_INFO "src  = %d\n",ntohs(tcph->source));
  printk(KERN_INFO "dest = %d\n",ntohs(tcph->dest));
  printk(KERN_INFO "doff = %d\n",tcph->doff);
  printk(KERN_INFO "res1 = %d\n",tcph->res1);
  printk(KERN_INFO "res2 = %d\n",tcph->res2);
  printk(KERN_INFO "urg  = %d\n",tcph->urg);
  printk(KERN_INFO "ack  = %d\n",tcph->ack);
  printk(KERN_INFO "psh  = %d\n",tcph->psh);
  printk(KERN_INFO "rst  = %d\n",tcph->rst);
  printk(KERN_INFO "syn  = %d\n",tcph->syn);
  printk(KERN_INFO "fin  = %d\n",tcph->fin);
  printk(KERN_INFO "wind = %d\n",ntohs(tcph->window));
*/
 }
 else
 {
  tcph = NULL;
 }

 iph = (struct iphdr*)skb_push(skb, sizeof(struct iphdr));
 len += sizeof(struct iphdr);
 *length+=len;
 *offset-=len;

 if ((ciph.version == LITE_TCP)
  || (ciph.version == LITE_TCP_COMP)
  || (ciph.version == LITE_UDP)
  || (ciph.version == LITE_ICMP))
 {
  memcpy(&iph->saddr,&ciph.saddr0,8);
  if (ciph.pair != NOTFOUND)
  {
   pair = lite_get_pair_from_addresses(&cipe->pairs, iph->saddr,
                                                 iph->daddr, LITE_RECEIVE);
   if (pair != NOTFOUND)
   {
    cipe->pairs.saddr[pair] = 0;
    cipe->pairs.daddr[pair] = 0;
    cipe->pairs.ok[pair] = 0;
   }
   if (ciph.pair < OKFLAG)
   {
    cipe->pairs.ok[ciph.pair] = UKNOW;
   }
   else
   {
    ciph.pair -= OKFLAG;
    cipe->pairs.ok[ciph.pair] = IKNOWUKNOW;
   }
   cipe->pairs.saddr[ciph.pair] = iph->daddr;
   cipe->pairs.daddr[ciph.pair] = iph->saddr;
  }
 }
 else
 {
  iph->daddr = INADDR_NONE;
  iph->saddr = INADDR_NONE;
  if (ciph.pair != NOTFOUND)
  {
   if (!(ciph.pair < OKFLAG))
   {
    ciph.pair -= OKFLAG;
   }
   if (cipe->pairs.ok[ciph.pair] > IKNOW)
   {
    cipe->pairs.ok[ciph.pair] = IKNOWUKNOW;
    iph->saddr = cipe->pairs.daddr[ciph.pair];
    iph->daddr = cipe->pairs.saddr[ciph.pair];
   }
  }
 }

/*
 if ((ciph.version == LITE_TCP_SHORT_COMP)
  || (ciph.version == LITE_TCP_COMP))
 {
  tcph->check  = 0;
  len = *length - sizeof(struct tcphdr);
  tcph->check  = tcp_check(tcph, sizeof(struct tcphdr), iph->saddr, iph->daddr,
                  block_crc((char*)(tcph + sizeof(struct tcphdr)), len));
  printk(KERN_INFO "our crc = %d, tcp crc = %d\n", tcph->check, ctcph.check);
  tcph->check  = ctcph.check;
 }
*/

 iph->ihl     = 5;
 iph->version = 4;
 iph->tos     = IPTOS_LOWDELAY;
 iph->tot_len = htons(*length);
 iph->id      = htons(ip_id_count++);
 iph->frag_off= htons(IP_DF);
 iph->ttl     = ttl;
 iph->protocol= proto;
 iph->check   = 0;

 ip_send_check(iph);
 return;

}

int litec(struct cipe* c, unsigned char* buf, int* len)
{
 unsigned char* lzo_w;
 unsigned char* lzo_b;
 int in_len;
 int out_len;

 lzo_w = c->lzo_wrk;
 lzo_b = c->lzo_buf;
 in_len = *len - 12;

 if (in_len > 350)
 {
  if (lzo_litec (buf+12, lzo_b, lzo_w, in_len, &out_len))
  {
   if (in_len > out_len)
   {
    *len = out_len + 12;
    memcpy(buf+12, lzo_b, out_len);
    return(in_len - out_len);
   }
  }
 }
 return(0);
}

int lited(struct cipe* c, unsigned char* buf, int* len)
{
 unsigned char* lzo_b;
 int out_len;

 lzo_b = c->lzo_buf;

 if (lzo_lited(buf+4, lzo_b, *len, &out_len))
 {
  memcpy(buf, lzo_b, out_len);
  *len = out_len;
  return(1);
 }
 return(0);
}

