/*
   CIPE - encrypted IP over UDP tunneling

   sock.c - socket/input interface

   Copyright 1996 Olaf Titz <olaf@bigred.inka.de>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version
   2 of the License, or (at your option) any later version.
*/
/* $Id: sock.c,v 1.11 1998/03/15 16:31:14 olaf Exp $ */

#include <net/ip.h>
#include "cipe.h"
#include <linux/if.h>
#include <net/sock.h>
#include <linux/module.h>
#include <linux/config.h>

#ifdef VER_LITE
#include "liteprototypes.h"
#endif

/* Rewire generic socket operations to our device-specific ones.
   We have new routines for close, sendmsg, recvmsg. */

/* init struct cipe *c based on struct sock *sk */
#define SOCKTOC(nam,sk,c,r) \
    struct cipe *c=(struct cipe *)sk->protinfo.af_packet.bound_dev; \
    if ((!c) || (c->magic!=CIPE_MAGIC)) { \
	 printk(KERN_ERR "Ouch: SOCKTOC" nam "\n"); return r; } \
    else { \
         dprintk1(DEB_CALL, KERN_INFO, "%s: " nam "\n", c->dev->name); }


/* Close the socket */
void cipe_sock_close(struct sock *sock, unsigned long timeout)
{
    SOCKTOC("sock_close",sock,c,);
    c->sock=NULL;

    /* Put back the old protocol block and let it do the close */
    sock->prot=c->udp_prot;
    sock->prot->close(sock, timeout);
    if (c->dev->flags&IFF_UP)
	dev_close(c->dev);
    else
	cipe_close(c);
    MOD_DEC_USE_COUNT;
}

/* Anything we send to this socket is in fact a key exchange block.
   Encode and encrypt it accordingly.
*/
int cipe_sendmsg(struct sock *sock, struct msghdr *msg,
		 int len, int nonblock, int flags)
{
    struct msghdr mymsg;
    struct iovec myio[2];
    struct sockaddr_in sa;
    struct sockshdr sh;
    int l, n=0;
    unsigned char buf[KEYXCHGBLKMAX];
    unsigned short fs;
    SOCKTOC("cipe_sendmsg",sock,c,-ENOSYS);

    if (len<KEYXCHGBLKMIN)
	return -EINVAL;
    if (len>KEYXCHGBLKMAX)
	return -EMSGSIZE;
    cipe_prnpad(buf, sizeof(buf));
#ifdef VER_SHORT
    memcpy_fromiovec(buf, msg->msg_iov, len);
#else
#ifdef VER_LITE
    memcpy_fromiovec(buf+blockSize+4, msg->msg_iov, len);
#else
    memcpy_fromiovec(buf+blockSize, msg->msg_iov, len);
#endif
#endif
    len=KEYXCHGBLKMIN+buf[sizeof(buf)-1]; /* random */
    cipe_encrypt(c, buf, &len, TW_NEWKEY);

    sa.sin_family=AF_INET;
    sa.sin_addr.s_addr=c->peeraddr;
    sa.sin_port=c->peerport;

    if (c->sockshost) {
	/* Prepend a socks header. */
	memset(&sh, 0, 4);
	sh.atyp=1;
	sh.dstaddr=c->sockshost;
	sh.dstport=c->socksport;
	myio[n].iov_base=&sh;
	myio[n].iov_len=sizeof(sh);
	++n;
    }
    myio[n].iov_base=&buf;
    myio[n].iov_len=len;
    /* mymsg=*msg; */
    mymsg.msg_name=&sa;
    mymsg.msg_namelen=sizeof(sa);
    mymsg.msg_iov=myio;
    mymsg.msg_iovlen=n+1;
    /* just to be sure */
    mymsg.msg_control=NULL;
    mymsg.msg_controllen=0;
    mymsg.msg_flags=0;

    /* Call the real thing. Pretend this is user space segment. */
    fs=get_fs();
    set_fs(get_ds());
    if (c->sockshost)
	len+=sizeof(struct sockshdr);
    dprintk3(DEB_KXC, KERN_INFO, "%s: real sendmsg len %d text=%04x...\n",
	     c->dev->name, len,
	     ntohs(*((unsigned short *)(myio[n].iov_base))));
    l=c->udp_prot->sendmsg(sock, &mymsg, len, nonblock, flags);
    set_fs(fs);
    return l;
}

/* Check if we have a new peer (socks relayer changed, etc) */
static void checkpeer(struct cipe *c, __u32 saddr, __u16 sport)
{
    if (c->peeraddr!=saddr || c->peerport!=sport) {
	unsigned char *x=(unsigned char *)&saddr;
	c->peeraddr=saddr;
	c->peerport=sport;
	printk(KERN_NOTICE "%s: new peer %d.%d.%d.%d:%d\n",
	       c->dev->name, x[0], x[1], x[2], x[3], ntohs(sport));
    }
}

/* Decrypt a received packet. Requeue it or return kxc block. */
/* On entry the packet starts with the original IP header (?),
   ip_hdr and h.uh are set to the IP and UDP headers. */
struct sk_buff *cipe_decrypt_skb(struct cipe *c, struct sock *sk,
				 struct sk_buff *skb, int *offset)
{
    struct sk_buff *n;
    int length;
    int tail;
    __u32 rsaddr;
    __u16 rsport;

#ifdef VER_LITE
    length = sizeof(struct iphdr) - PIPHDRLEN
           + sizeof(struct tcphdr) - CTCPHDRLEN;
    tail = 3600;
#else
    length = 0;
    tail = 0;
#endif
    n=alloc_skb(skb->len + length + tail, GFP_KERNEL);
    if (!n) {
	printk(KERN_WARNING "%s: cipe_decrypt_skb(): out of memory\n",
	       c->dev->name);
	++c->stat.rx_dropped;
	return NULL;
    }
    n->sk=NULL;
    n->free=1;
    n->dev=c->dev;
    n->len=skb->len+length+tail;
    skb_pull(n, length);

    length=ntohs(skb->h.uh->len)-sizeof(struct udphdr);

    /* Copy the datagram into new buffer, skip IP header */
    /* We must copy here because raw sockets get only a clone of the
       skbuff, so they would receive the plaintext */

    rsaddr=skb->daddr;
    rsport=skb->h.uh->source;
    n->len=skb->len+tail;
    memcpy(n->data, skb->h.raw, skb->len);
    /*n->ip_hdr=NULL;*/
    n->h.uh=(struct udphdr *)n->data;
    dprintk4(DEB_INP, KERN_INFO, "UDP: s=%d d=%d l=%d c=%04X\n",
	     ntohs(n->h.uh->source), ntohs(n->h.uh->dest),
	     ntohs(n->h.uh->len), n->h.uh->check);

    if (c->sockshost) {
	/* Pull out the SOCKS header and correct the peer's address. */
	struct sockshdr *sh;
	sh=(struct sockshdr *)skb_pull(n, sizeof(struct udphdr));
	dprintk4(DEB_INP, KERN_INFO, "socks: fr=%d at=%d da=%08lX dp=%d\n",
		 sh->frag, sh->atyp, ntohl(sh->dstaddr), ntohs(sh->dstport));
	if (sh->frag || (sh->atyp!=1))
	    goto error;
	/* Pull out UDP header */
	n->ip_hdr=(struct iphdr *)skb_pull(n, sizeof(struct sockshdr));
	n->saddr=sh->dstaddr;
	n->h.uh->source=sh->dstport;
	length-=sizeof(struct sockshdr);
	*offset=sizeof(struct sockshdr)+sizeof(struct udphdr);
    } else {
	/* Pull out UDP header */
	n->ip_hdr=(struct iphdr *) skb_pull(n, sizeof(struct udphdr));
	n->saddr=rsaddr;
	n->h.uh->source=rsport;
	*offset=sizeof(struct udphdr);
    }

    if (c->havekey) {
        switch (cipe_decrypt(c, n->data, &length)) {
	case TW_DATA:
	    break;
	case TW_NEWKEY:
	    /* return it as key exchange block - proper UDP datagram */
            skb_trim(n, length);
	    checkpeer(c, rsaddr, rsport);
#if 0
	    n->saddr=c->myaddr;
	    n->daddr=c->peeraddr;
#endif
	    n->h.uh->check=0;

#ifdef VER_LITE
            memset(&c->pairs, 0, sizeof(struct pairs));
	    printk(KERN_INFO "%s: rx other          : %lu \n",
	       c->dev->name,
               c->rx_lite_other);
	    printk(KERN_INFO "%s: rx icmp           : %lu \n",
	       c->dev->name,
               c->rx_lite_icmp);
	    printk(KERN_INFO "%s: rx icmp,short     : %lu \n",
	       c->dev->name,
               c->rx_lite_icmp_short);
	    printk(KERN_INFO "%s: rx udp            : %lu \n",
	       c->dev->name,
               c->rx_lite_udp);
	    printk(KERN_INFO "%s: rx udp,short      : %lu \n",
	       c->dev->name,
               c->rx_lite_udp_short);
	    printk(KERN_INFO "%s: rx tcp            : %lu \n",
	       c->dev->name,
               c->rx_lite_tcp);
	    printk(KERN_INFO "%s: rx tcp,short      : %lu \n",
	       c->dev->name,
               c->rx_lite_tcp_short);
	    printk(KERN_INFO "%s: rx tcp,comp       : %lu \n",
	       c->dev->name,
               c->rx_lite_tcp_comp);
	    printk(KERN_INFO "%s: rx tcp,short,comp : %lu \n",
	       c->dev->name,
               c->rx_lite_tcp_short_comp);
	    printk(KERN_INFO "%s: rx uncompr. pkts  : %lu \n",
	       c->dev->name,
               c->rx_uncomp_pkts);
	    printk(KERN_INFO "%s: rx uncompr. bytes : %lu \n",
	       c->dev->name,
               c->rx_uncomp_bytes);
	    printk(KERN_INFO "%s: rx compr. pkts    : %lu \n",
	       c->dev->name,
               c->rx_comp_pkts);
	    printk(KERN_INFO "%s: rx comp. raw bytes: %lu \n",
	       c->dev->name,
               c->rx_comp_rbytes);
	    printk(KERN_INFO "%s: rx comp. rx bytes : %lu \n",
	       c->dev->name,
               c->rx_comp_cbytes);
	    printk(KERN_INFO "%s: rx with dyn key   : %lu \n",
	       c->dev->name,
               c->rx_dkey);
	    printk(KERN_INFO "%s: rx with fixed key : %lu \n",
	       c->dev->name,
               c->rx_fkey);
	    printk(KERN_INFO "%s: tx other          : %lu \n",
	       c->dev->name,
               c->tx_lite_other);
	    printk(KERN_INFO "%s: tx icmp           : %lu \n",
	       c->dev->name,
               c->tx_lite_icmp);
	    printk(KERN_INFO "%s: tx icmp,short     : %lu \n",
	       c->dev->name,
               c->tx_lite_icmp_short);
	    printk(KERN_INFO "%s: tx udp            : %lu \n",
	       c->dev->name,
               c->tx_lite_udp);
	    printk(KERN_INFO "%s: tx udp,short      : %lu \n",
	       c->dev->name,
               c->tx_lite_udp_short);
	    printk(KERN_INFO "%s: tx tcp            : %lu \n",
	       c->dev->name,
               c->tx_lite_tcp);
	    printk(KERN_INFO "%s: tx tcp,short      : %lu \n",
	       c->dev->name,
               c->tx_lite_tcp_short);
	    printk(KERN_INFO "%s: tx tcp,comp       : %lu \n",
	       c->dev->name,
               c->tx_lite_tcp_comp);
	    printk(KERN_INFO "%s: tx tcp,short,comp : %lu \n",
	       c->dev->name,
               c->tx_lite_tcp_short_comp);
	    printk(KERN_INFO "%s: tx uncompr. pkts  : %lu \n",
	       c->dev->name,
               c->tx_uncomp_pkts);
	    printk(KERN_INFO "%s: tx uncompr. bytes : %lu \n",
	       c->dev->name,
               c->tx_uncomp_bytes);
	    printk(KERN_INFO "%s: tx compr. pkts    : %lu \n",
	       c->dev->name,
               c->tx_comp_pkts);
	    printk(KERN_INFO "%s: tx comp. raw bytes: %lu \n",
	       c->dev->name,
               c->tx_comp_rbytes);
	    printk(KERN_INFO "%s: tx comp. tx bytes : %lu \n",
	       c->dev->name,
               c->tx_comp_cbytes);
	    printk(KERN_INFO "%s: tx with dyn key   : %lu \n",
	       c->dev->name,
               c->tx_dkey);
	    printk(KERN_INFO "%s: tx with fixed key : %lu \n",
	       c->dev->name,
               c->tx_fkey);
#endif
	    return n;
	case TW_CTRL:
	    /* process a control packet - to be implemented */
	    checkpeer(c, rsaddr, rsport);
	default:
	    /* Bogus packet etc. */
	    ++c->stat.rx_crc_errors;
	    goto error;
	}
    } else if (!c->mayclear) {
	goto error;
    }

#ifdef VER_LITE
    litex(c, n, &length, offset);
#endif    

    skb_trim(n, length);
    checkpeer(c, rsaddr, rsport);
    n->h.iph=(struct iphdr *)n->data;
    n->mac.raw=n->data; /* no hardware header */
    memset(n->proto_priv, 0, sizeof(struct options));
    n->protocol = htons(ETH_P_IP);
    n->ip_summed = 0;

#ifdef DEBUG
    if (cipe_debug&DEB_INP) {
	unsigned char *x=(unsigned char *)(&n->h.iph->tos - 1);
	printk(KERN_INFO
	       "%s: %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x length %d\n",
	       n->dev->name,
	       x[0], x[1], x[2], x[3], x[4], x[5], x[6], x[7], x[8], x[9],
	       x[10], x[11], x[12], x[13], x[14], x[15], x[16], x[17], x[18], x[19],
length);
    }
#endif

    /* requeue */
    netif_rx(n);
    ++c->stat.rx_packets;
    return NULL;

#if 0
 framerr:
    ++c->stat.rx_frame_errors; /* slightly abuse this */
#endif
 error:
    ++c->stat.rx_errors;
    /*n->sk=NULL;*/
    kfree_skb(n, FREE_READ);
    return NULL;
}

/* Receive message. If encrypted, put it through the mill.
   If decrypted, return it as key exchange block.
   This is mostly from net/ipv4/udp.c, but cipe_decrypt_skb pulls the
   UDP header.
*/

int cipe_recvmsg(struct sock *sk, struct msghdr *msg, int len,
		 int noblock, int flags,int *addr_len)
{
  	int copied = 0;
  	int truesize;
  	struct sk_buff *skb, *skn;
  	int er, of;
  	struct sockaddr_in *sin=(struct sockaddr_in *)msg->msg_name;
	SOCKTOC("cipe_recvmsg",sk,c,-ENOSYS);

	/*
	 *	Check any passed addresses
	 */

  	if (addr_len)
  		*addr_len=sizeof(*sin);

	/*
	 *	From here the generic datagram does a lot of the work. Come
	 *	the finished NET3, it will do _ALL_ the work!
	 */

	do {
	    /* CIPE: look if the packet is encrypted, repeat until
	       a decrypted one is found */
	    skn=NULL;
	    skb=skb_recv_datagram(sk,flags,noblock,&er);
	    if(skb==NULL) {
		dprintk2(DEB_KXC, KERN_INFO, "%s: skb_recv_datagram: %d\n",
			 c->dev->name, er);
  		return er;
	    }

	    if (!skb->h.uh->source) {
		/* Synthetic KXC packets are marked by source port 0.
		   Correct this - we know the truth from our structure.
		   Perhaps it would be better to not correct it so the
		   user level daemon can spot the difference? */
		skb->h.uh->source=c->peerport;
		of=skb->data-skb->head+sizeof(struct udphdr);
		break;
	    }
	    skn=cipe_decrypt_skb(c, sk, skb, &of);
	    skb_free_datagram(sk, skb);
	    skb=skn;
	} while (skb==NULL);

  	truesize = skb->len;
  	copied = MIN(len, truesize);

  	/*
  	 *	FIXME : should use udp header size info value
  	 */
	skb_copy_datagram_iovec(skb,of,msg->msg_iov,copied);
	sk->stamp=skb->stamp;

	/* Copy the address. */
	if (sin)
	{
		sin->sin_family = AF_INET;
		sin->sin_port = skb->h.uh->source;
		sin->sin_addr.s_addr = skb->daddr;
#ifdef CONFIG_IP_TRANSPARENT_PROXY
		/* Does this ever apply here? */
		if (flags&MSG_PROXY)
		{
			/*
			 * We map the first 8 bytes of a second sockaddr_in
			 * into the last 8 (unused) bytes of a sockaddr_in.
			 * This _is_ ugly, but it's the only way to do it
			 * easily,  without adding system calls.
			 */
			struct sockaddr_in *sinto =
				(struct sockaddr_in *) sin->sin_zero;

			sinto->sin_family = AF_INET;
			sinto->sin_port = skb->h.uh->dest;
			sinto->sin_addr.s_addr = skb->saddr;
		}
#endif
	}

	if (skb==skn)
	    /* the buffer was a copy made by decryption */
	    kfree_skb(skb, FREE_READ);
	else
	    skb_free_datagram(sk, skb);

  	return(copied);
}

#define sreturn(x) {sti(); return((x));}

int cipe_attach(struct device *dev, struct siocsifcipatt *parm)
{
    struct file *file;
    struct inode *inode;
    struct socket *sock;
    struct sock *sk;
    int fd=parm->fd;
    struct cipe *c0;
    DEVTOCIPE(dev,c,-ENODEV);
    cli();
    if (c->sock)
	sreturn(-EBUSY);

    /* Find the socket (from net/socket.c) */
    if (fd < 0 || fd >= NR_OPEN || !(file = current->files->fd[fd]))
	sreturn(-EBADF);
    inode = file->f_inode;
    if (!inode || !inode->i_sock)
	sreturn(-ENOTSOCK);
    sock=&inode->u.socket_i;
    if (sock->type!=SOCK_DGRAM)
	sreturn(-ESOCKTNOSUPPORT);
    if (sock->ops->family!=AF_INET)
	sreturn(-EAFNOSUPPORT);
    sk=(struct sock *)sock->data;
    if (sk->state!=TCP_ESTABLISHED)
	sreturn(-ENOTCONN);
    if (((c0=(struct cipe *)sk->protinfo.af_packet.bound_dev)) &&
	(c0->magic==CIPE_MAGIC))
	sreturn(-EBUSY); /* socket is already attached */

    MOD_INC_USE_COUNT;
    c->sock=sk;
    c->peeraddr=sk->daddr;
    c->peerport=sk->dummy_th.dest;
    c->myaddr=sk->saddr;
    c->myport=sk->dummy_th.source;
    /* Disconnect socket, we might receive from somewhere else */
    sk->daddr=0;
    sk->dummy_th.dest=0;

    /* Set up new socket operations */
    c->udp_prot=sk->prot;
    c->cipe_proto=*sk->prot;
    c->cipe_proto.close=cipe_sock_close;
    c->cipe_proto.sendmsg=cipe_sendmsg;
    c->cipe_proto.recvmsg=cipe_recvmsg;
    sk->prot=&c->cipe_proto;

    /* Fill an otherwise unused field in the sock struct with this info.
       Actually, this is very similar to a packet socket!
       The ugly cast saves us one deref in the actual ops */
    sk->protinfo.af_packet.bound_dev=(struct device *)c;
    sk->no_check=1; /* our packets are checksummed internally */

    sti();
    return 0;
}

/* Build and enqueue a fake UDP packet to receive.
   Note that these are neither encrypted nor SOCKSed.
*/
void cipe_fakenkey(struct cipe *c, char typ)
{
    int len=sizeof(struct udphdr)+1;
    struct sk_buff *skb=alloc_skb(len, GFP_ATOMIC);

    if (!skb) {
        printk(KERN_WARNING "%s: cipe_fakenkey(): out of memory\n",
               c->dev->name);
        return; /* not much we can do */
    }

    dprintk2(DEB_KXC, KERN_INFO, "%s: fake kxc block typ=%d\n",
             c->dev->name, typ);

    skb->sk=NULL;
    skb->free=1; /* Discard after use */
    skb->ip_hdr=NULL;
    skb->h.uh=(struct udphdr *)skb->data;
    skb->len=len;
    skb->saddr=c->myaddr;
    skb->daddr=c->peeraddr;

    skb->h.uh->source=0; /* mark as KXC packet */
    skb->h.uh->dest=c->myport;
    len-=sizeof(struct udphdr);
    skb->h.uh->len=len;
    skb->h.uh->check=0;
    skb->mac.raw=skb->data; /* no hardware header */

    /* All those contortions for just one byte of payload data.
       Since we generate only NK_RREQ and NK_REQ it's effectively
       one _bit_... */
    skb->data[sizeof(struct udphdr)]=typ;

    if (sock_queue_rcv_skb(c->sock, skb)<0) {
        printk(KERN_WARNING "%s: cipe_fakenkey(): enqueuing failed\n",
               c->dev->name);
        kfree_skb(skb, FREE_WRITE);
    }
}
